import logging

from django.conf import settings
from django.core.management import BaseCommand

from rele import config
from rele.management.discover import discover_subs_modules
from rele.worker import create_and_run

logger = logging.getLogger(__name__)


class Command(BaseCommand):
    help = "Start subscriber threads to consume messages from Relé topics."
    config = config.Config(settings.RELE)

    def handle(self, *args, **options):
        if all(map(lambda x: x.get("CONN_MAX_AGE"), settings.DATABASES.values())):
            self.stderr.write(
                self.style.WARNING(
                    "WARNING: settings.CONN_MAX_AGE is not set to 0. "
                    "This may result in slots for database connections to "
                    "be exhausted."
                )
            )
        subs = config.load_subscriptions_from_paths(
            discover_subs_modules(), self.config.sub_prefix, self.config.filter_by
        )
        self.stdout.write(f"Configuring worker with {len(subs)} " f"subscription(s)...")
        create_and_run(subs, self.config)



import importlib
import warnings

_middlewares = []

default_middleware = ["rele.contrib.LoggingMiddleware"]
DEPRECATED_HOOKS = ["post_publish"]


def register_middleware(config, **kwargs):
    paths = config.middleware
    global _middlewares
    _middlewares = []
    for path in paths:
        *module_parts, middleware_class = path.split(".")
        module_path = ".".join(module_parts)
        module = importlib.import_module(module_path)
        middleware = getattr(module, middleware_class)()
        middleware.setup(config, **kwargs)
        _middlewares.append(middleware)


def run_middleware_hook(hook_name, *args, **kwargs):
    for middleware in _middlewares:
        if hook_name not in DEPRECATED_HOOKS or hasattr(middleware, hook_name):
            getattr(middleware, hook_name)(*args, **kwargs)


class WarnDeprecatedHooks(type):
    def __new__(cls, *args, **kwargs):
        x = super().__new__(cls, *args, **kwargs)
        for deprecated_hook in DEPRECATED_HOOKS:
            if hasattr(x, deprecated_hook):
                warnings.warn(
                    "The post_publish hook in the middleware is deprecated "
                    "and will be removed in future versions. Please substitute it with "
                    "the post_publish_success hook instead.",
                    DeprecationWarning,
                )
        return x


class BaseMiddleware(metaclass=WarnDeprecatedHooks):
    """Base class for middleware.  The default implementations
    for all hooks are no-ops and subclasses may implement whatever
    subset of hooks they like.
    """

    def setup(self, config, **kwargs):
        """Called when middleware is registered.
        :param config: Relé Config object
        """

    def pre_publish(self, topic, data, attrs):
        """Called before Publisher sends message.
        :param topic:
        :param data:
        :param attrs:
        """

    def post_publish_success(self, topic, data, attrs):
        """Called after Publisher succesfully sends message.
        :param topic:
        :param data:
        :param attrs:
        """

    def post_publish_failure(self, topic, exception, message):
        """Called after publishing fails.
        :param topic:
        :param exception:
        :param message:
        """

    def pre_process_message(self, subscription, message):
        """Called when the Worker receives a message.
        :param subscription:
        :param message:
        """

    def post_process_message(self):
        """Called after the Worker processes the message."""

    def post_process_message_success(self, subscription, start_time, message):
        """Called after the message has been successfully processed.
        :param subscription:
        :param start_time:
        :param message:
        """

    def post_process_message_failure(
        self, subscription, exception, start_time, message
    ):
        """Called after the message has been unsuccessfully processed.
        :param subscription:
        :param exception:
        :param start_time:
        :param message:
        """

    def pre_worker_start(self):
        """Called before the Worker process starts up."""

    def post_worker_start(self):
        """Called after the Worker process starts up."""

    def pre_worker_stop(self, subscriptions):
        """Called before the Worker process shuts down."""

    def post_worker_stop(self):
        """Called after the Worker process shuts down."""



import json

from rele.contrib.logging_middleware import LoggingMiddleware


class VerboseLoggingMiddleware(LoggingMiddleware):
    def setup(self, config, **kwargs):
        super().setup(config, **kwargs)
        self._encoder = config.encoder

    def post_publish_success(self, topic, message_data, message_attributes):
        self._logger.info(
            f"Successfully published to {topic}",
            extra={
                "pubsub_publisher_attrs": message_attributes,
                "metrics": {
                    "name": "publications",
                    "data": {"agent": self._app_name, "topic": topic},
                },
                "subscription_message": json.dumps(message_data, cls=self._encoder),
            },
        )

    def post_process_message_success(self, subscription, start_time, message):
        self._logger.info(
            f"Successfully processed message for {subscription}",
            extra={
                "metrics": {
                    "name": "subscriptions",
                    "data": self._build_data_metrics(
                        subscription, message, "succeeded", start_time
                    ),
                },
                "subscription_message": str(_VerboseMessage(message)),
            },
        )

    def post_process_message_failure(
        self, subscription, exception, start_time, message
    ):
        super().post_process_message_failure(
            subscription, exception, start_time, _VerboseMessage(message)
        )


class _VerboseMessage:
    def __init__(self, message):
        self._message = message
        self.attributes = message.attributes

    def __repr__(self):
        _MESSAGE_REPR = """\
Message {{
  data: {!r}
  ordering_key: {!r}
  attributes: {}
}}"""

        data = self._message._message.data
        attrs = self._message_attrs_repr()
        ordering_key = str(self._message.ordering_key)

        return _MESSAGE_REPR.format(data, ordering_key, attrs)

    def _message_attrs_repr(self):
        message_attrs = json.dumps(
            dict(self.attributes), indent=2, separators=(",", ": "), sort_keys=True
        )

        indented = []
        for line in message_attrs.split("\n"):
            indented.append("  " + line)

        message_attrs = "\n".join(indented)
        message_attrs = message_attrs.lstrip()

        return message_attrs



import argparse
import importlib
import logging
import os
import sys

from rele import config, discover
from rele.worker import create_and_run

logger = logging.getLogger(__name__)


def main():
    # modify path so we can import modules and packages
    sys.path.insert(0, os.getcwd())

    parser = argparse.ArgumentParser(
        prog="Relé", description="Harness the power of Relé from the command line"
    )

    subparsers = parser.add_subparsers(help="Select a command", dest="command")

    run_parser = subparsers.add_parser(
        "run",
        help="Run a Relé worker with auto-discovery of subs modules in the "
        "current path. Auto-discovery will include all subs "
        "and settings modules. If no settings module is discovered, "
        "defaults will be used.",
    )
    run_parser.add_argument(
        "--settings",
        "-s",
        default=None,
        required=False,
        help="Settings file dot path. Ex. project.settings. "
        "If none is supplied, Relé will attempt to autodiscover in the root path.",
    )
    run_parser.add_argument(
        "--third-party-subscriptions",
        default=None,
        required=False,
        nargs="+",
        help="Specify subscriptions from the third-party packages. "
        "Example --third-party-subscriptions my_package.subs another_package.subs",
    )
    args = parser.parse_args()

    if args.command == "run":
        run_worker(args.settings, args.third_party_subscriptions)


def run_worker(settings, third_party_subs):
    settings, module_paths = discover.sub_modules(settings)

    if third_party_subs:
        validated_third_party_subs = _validated_third_party_subs(third_party_subs)
        module_paths = module_paths + validated_third_party_subs

    configuration = config.setup(settings.RELE if settings else None)
    subs = config.load_subscriptions_from_paths(
        module_paths, configuration.sub_prefix, configuration.filter_by
    )
    create_and_run(subs, configuration)


def _validated_third_party_subs(subs_import_path):
    validated_subs = []
    for sub_path in subs_import_path:
        if _is_valid_module(sub_path):
            validated_subs.append(sub_path)

    return validated_subs


def _is_valid_module(import_path):
    is_valid = False
    try:
        importlib.import_module(import_path)
        is_valid = True
    except ModuleNotFoundError:
        pass

    return is_valid



from dataclasses import dataclass


@dataclass
class RetryPolicy:
    """A RetryPolicy object encapsulates the validation rules.

    Defines retry policy settings and ensures the values are correct.
    If provided values are wrong, a ValidationError is raised.

    :param minimum_backoff: int Accepts values greater than 0
    :param maximum_backoff: int Accepts values greater than minimum_backoff.
    """

    minimum_backoff: int
    maximum_backoff: int

    def __init__(self, minimum_backoff, maximum_backoff):
        self._guard_against_wrong_parameters(minimum_backoff, maximum_backoff)

        self.minimum_backoff = minimum_backoff
        self.maximum_backoff = maximum_backoff

    def _guard_against_wrong_parameters(self, minimum_backoff, maximum_backoff):
        if minimum_backoff == 0:
            raise ValueError("minimum_backoff must be greater than 0")

        if maximum_backoff == 0:
            raise ValueError("maximum_backoff must be greater than 0")

        if minimum_backoff > maximum_backoff:
            raise ValueError("minimum_backoff should be less than maximum_backoff.")



__version__ = "1.15.0"

try:
    import django
except ImportError:
    pass
else:
    if django.__version__ < "3.2":
        default_app_config = "rele.apps.ReleConfig"

from .client import Publisher, Subscriber  # noqa
from .config import setup  # noqa
from .publishing import publish  # noqa
from .subscription import Callback, Subscription, sub  # noqa
from .worker import Worker  # noqa



from .flask_middleware import FlaskMiddleware  # noqa
from .logging_middleware import LoggingMiddleware  # noqa
from .unrecoverable_middleware import UnrecoverableMiddleWare  # noqa
from .verbose_logging_middleware import VerboseLoggingMiddleware  # noqa

try:
    from .django_db_middleware import DjangoDBMiddleware  # noqa
except ImportError:
    pass
